package lcsb.mapviewer.services.search.mirna;

import java.util.ArrayList;
import java.util.Collection;
import java.util.Collections;
import java.util.HashSet;
import java.util.List;
import java.util.Set;

import org.apache.logging.log4j.LogManager;
import org.apache.logging.log4j.Logger;
import org.springframework.stereotype.Service;
import org.springframework.transaction.annotation.Transactional;

import lcsb.mapviewer.annotation.data.MiRNA;
import lcsb.mapviewer.annotation.data.Target;
import lcsb.mapviewer.annotation.services.MiRNAParser;
import lcsb.mapviewer.annotation.services.MiRNASearchException;
import lcsb.mapviewer.annotation.services.annotators.AnnotatorException;
import lcsb.mapviewer.annotation.services.annotators.HgncAnnotator;
import lcsb.mapviewer.common.IProgressUpdater;
import lcsb.mapviewer.model.Project;
import lcsb.mapviewer.model.map.BioEntity;
import lcsb.mapviewer.model.map.MiriamData;
import lcsb.mapviewer.model.map.MiriamType;
import lcsb.mapviewer.model.map.model.Model;
import lcsb.mapviewer.model.map.species.Element;
import lcsb.mapviewer.model.map.species.Gene;
import lcsb.mapviewer.model.map.species.Protein;
import lcsb.mapviewer.model.map.species.Rna;
import lcsb.mapviewer.services.search.DbSearchCriteria;
import lcsb.mapviewer.services.search.DbSearchService;

/**
 * Implementation of the service that allows access information about miRNAs.
 * 
 * @author Ayan Rota
 * 
 */
@Transactional
@Service
public class MiRNAService extends DbSearchService implements IMiRNAService {

  /**
   * Default class logger.
   */
  private Logger logger = LogManager.getLogger();

  /**
   * Access point and parser for the online ctd database.
   */
  private MiRNAParser miRNAParser;

  /**
   * Object used to get information about {@link MiriamType#HGNC} annotations and
   * transformations.
   */
  private HgncAnnotator hgncAnnotator;
  
  public MiRNAService(final MiRNAParser miRNAParser, final HgncAnnotator hgncAnnotator) {
    super();
    this.miRNAParser = miRNAParser;
    this.hgncAnnotator = hgncAnnotator;
  }

  @Override
  public MiRNA getByName(final String name, final DbSearchCriteria searchCriteria) {
    if ("".equals(name)) {
      return null;
    }
    try {
      List<String> names = new ArrayList<>();
      names.add(name);
      List<MiRNA> miRNAs = miRNAParser.getMiRnasByNames(names);
      if (miRNAs.size() > 0) {
        removeUnknownOrganisms(miRNAs.get(0), searchCriteria.getOrganisms());
        super.addProjectTargets(searchCriteria.getProject(), miRNAs.get(0));
        return miRNAs.get(0);
      }
      return null;
    } catch (final MiRNASearchException e) {
      logger.error("Problem with accessing mirna database", e);
      return null;
    }
  }

  @Override
  public List<MiRNA> getForTargets(final Collection<Element> targets, final DbSearchCriteria searchCriteria) {
    List<MiRNA> mirnaList = new ArrayList<MiRNA>();
    Set<MiriamData> targetsMiriam = new HashSet<MiriamData>();
    for (final Element element : targets) {
      if (element instanceof Protein || element instanceof Gene || element instanceof Rna) {
        boolean hgncFound = false;
        for (final MiriamData md : element.getMiriamData()) {
          if (MiriamType.HGNC_SYMBOL.equals(md.getDataType())) {
            targetsMiriam.add(md);
            hgncFound = true;
          } else if (MiriamType.ENTREZ.equals(md.getDataType())) {
            // ad also entrez in case of mouse, rat, etc
            targetsMiriam.add(md);
          }

        }
        if (!hgncFound) {
          MiriamData md = new MiriamData(MiriamType.HGNC_SYMBOL, element.getName());
          try {
            if (hgncAnnotator.isValidHgncMiriam(md)) {
              targetsMiriam.add(md);
            }
          } catch (final AnnotatorException e) {
            logger.error("Problem with accessing HGNC database", e);
          }
        }
      }
    }
    try {
      mirnaList = miRNAParser.getMiRnaListByTargets(targetsMiriam);
    } catch (final MiRNASearchException e) {
      logger.error("Problem with accessing mirna database", e);
    }

    for (final MiRNA mirna : mirnaList) {
      removeUnknownOrganisms(mirna, searchCriteria.getOrganisms());
      super.addProjectTargets(searchCriteria.getProject(), mirna);
    }

    Collections.sort(mirnaList, new MiRNA.NameComparator());
    return mirnaList;
  }

  @Override
  public void cacheDataForModel(final Model originalModel, final IProgressUpdater progressUpdater) {
    logger.debug("Caching mirna queries...");
    double progress = 0.0;
    progressUpdater.setProgress(progress * IProgressUpdater.MAX_PROGRESS);
    List<Model> models = new ArrayList<>();
    models.add(originalModel);
    models.addAll(originalModel.getSubmodels());
    Set<MiriamData> targetMiriams = getHgncIdentifiers(models);
    targetMiriams.addAll(getEntrezIdentifiers(models));
    double counter = 0.0;
    int step = Math.max(targetMiriams.size() / 100, 1);
    List<MiriamData> currentList = new ArrayList<>();
    for (final MiriamData md : targetMiriams) {
      counter += 1;
      currentList.add(md);
      if (currentList.size() >= step) {
        logger.debug("check mirna target: " + currentList);
        try {
          List<MiRNA> mirnaList = miRNAParser.getMiRnaListByTargets(currentList);
          for (final MiRNA chemical : mirnaList) {
            cacheMiriamData(chemical);
          }
        } catch (final MiRNASearchException | AnnotatorException e) {
          logger.error("Problem with accessing info about mirna for target: " + md, e);
        }
        progress = counter / (double) targetMiriams.size();
        progressUpdater.setProgress(progress * IProgressUpdater.MAX_PROGRESS);
        currentList.clear();
      }
    }
    try {
      getSuggestedQueryList(originalModel.getProject());
    } catch (final MiRNASearchException e) {
      logger.error("Problem with accessing mirna info", e);
    }
  }

  private Set<MiriamData> getEntrezIdentifiers(final List<Model> models) {
    Set<MiriamData> targetMiriams = new HashSet<>();
    for (final Model model : models) {
      for (final BioEntity element : model.getBioEntities()) {
        for (final MiriamData md : element.getMiriamData()) {
          if (MiriamType.ENTREZ.equals(md.getDataType())) {
            targetMiriams.add(md);
          }
        }
      }
    }
    return targetMiriams;
  }

  /**
   * Removes targets for unknown organisms from the mirna.
   *
   * @param mirna
   *          mirna from which we want to remove targets
   * @param organisms
   *          organisms that should be kept
   */
  private void removeUnknownOrganisms(final MiRNA mirna, final List<MiriamData> organisms) {
    if (organisms.size() > 0) {
      List<Target> toRemove = new ArrayList<Target>();
      for (final Target target : mirna.getTargets()) {
        boolean remove = true;
        for (final MiriamData organism : organisms) {
          if (target.getOrganism() == null) {
            remove = false;
          } else if (target.getOrganism().equals(organism)) {
            remove = false;
          }
        }
        if (remove) {
          logger.debug("Target " + target.getName() + " removed from list because results are limited to organisms: "
              + organisms);
          toRemove.add(target);
        }
      }
      mirna.getTargets().removeAll(toRemove);
    }
  }

  @Override
  public List<String> getSuggestedQueryList(final Project project) throws MiRNASearchException {
    return miRNAParser.getSuggestedQueryList(project);
  }

}
